from .class_based_view import class_based_view, ViewBase
from .class_based_view import api_route, Route
from .class_based_view import fastapi_compatible_method

from .class_based_crud import class_based_crud_view, CrudView
